import subprocess
import struct
import random
import os
import hashlib

def _hash(x):
  return hashlib.sha256(repr(x).encode()).hexdigest()[:8]
def print_hash(name, x):
  print("{:<32}: {}".format(name, _hash(x)))

def parse_binary(b):
  return ''.join(map(lambda x:format(int(x), '0>2x'), b))

def print_header(s):
  l = 80
  print('{}\n{:^80}\n{}\n'.format(l * '_', s, l * '_'))

def parse_frame(buf):
  c = [0]
  def b(c, n):
    ret = buf[c[0]:c[0]+n]
    c[0] += n
    return ret
  dst_mac             = b(c, 6)
  src_mac             = b(c, 6)
  ether_type          = b(c, 2)
  ecpri_common_header = b(c, 4)
  pc_id               = b(c, 2)
  seq_id              = b(c, 2)
  iq_samples          = b(c, 240)

  avg = sum(map(lambda x:int(x), iq_samples))
  s = '''
| {:^12} | {:^12} | {:^12} | {:^12} | {:^12} | {:^12} |
| {:^12} | {:^12} | {:^12} | {:^12} | {:^12} | {:^12} |
| {} |'''.format(
              "DST", "SRC", "ETHER TYPE", "ECPRI HEADER", "PC ID", "SEQ ID",
              parse_binary(dst_mac),
              parse_binary(src_mac),
              parse_binary(ether_type),
              parse_binary(ecpri_common_header),
              parse_binary(pc_id),
              parse_binary(seq_id),
              parse_binary(iq_samples))[1:]
  return (avg == 0, s)

def read_trace(name, n):
  log_directory = '/root/ecpri-logs'
  file_name = '{}/{}'.format(log_directory, name)

  n = min(n * 262, os.path.getsize(file_name))
  f = open(file_name, "rb")
  data = f.read(n)
  f.close()

  print(name + ", frame number = " + str(n / 262))

  return data

def print_frame(data):
  n = len(data)
  frame_len = 262

  for i in range(int(n / frame_len)):
    empty, s = parse_frame(data[i * frame_len:(i+1) * frame_len])
    if empty:
      continue
    print("Frame example:\n\n{}\n".format(s))
    break

def analyze_tdd(data):
  null_frames = 0
  tdd_switch_list = []
  prev_empty = False
  prev_i = 0
  n = len(data)
  frame_len = 262

  for i in range(int(n / frame_len)):
    empty, s = parse_frame(data[i * frame_len:(i+1) * frame_len])
    if i > 0 and empty != prev_empty:
        tdd_switch_list.append(i-prev_i)
        prev_i = i
    null_frames += empty
    prev_empty = empty
  tdd_switch_list.append(i-prev_i)

  print('Frame analyzed: ' + str(i))

  total_frames = int(n / frame_len)
  ratio = total_frames / null_frames if null_frames > 0 else 'inf'
  print('TDD ratio: {}'.format(ratio))
  print('TDD switch list: ' + ', '.join(map(str, tdd_switch_list)) + '\n')

def print_frame_list(data, start, end):
  n = len(data)
  frame_len = 262

  end = int(n / frame_len) if end == -1 else end
  end = min(int(n / frame_len), end)

  for i in range(start, end):
    empty, s = parse_frame(data[i * frame_len:(i+1) * frame_len])
    print(s)
    print('')

def print_iq_list(data, start, end, n_channel=4):
  n = len(data)
  frame_len = 64 * n_channel

  end = int(n / frame_len) if end == -1 else end
  end = min(int(n / frame_len), end)

  for i in range(start * 4, end * 4, 4):
    iq = data[i * frame_len:(i+1) * frame_len]
    iq_packed = []
    k = 0

    h = _hash(iq)
    prev_x = 2
    for x in map(lambda x: int(x != 0), iq):
      if x == prev_x:
        iq_packed[-1][0] += 1
      else:
        iq_packed.append([1, str(x) * 2])
      prev_x = x
    print(h + "  " + " ".join(map(lambda x: "{}*{}".format(*x), iq_packed)))

def read_trx_trace(name, n, channels):
  log_directory = '/root/ecpri-logs'
  file_name = '{}/{}'.format(log_directory, name)

  n = min(n * (4 * channels * 64), os.path.getsize(file_name))
  f = open(file_name, "rb")
  data = f.read(n)
  f.close()

  print(name + ", frame number = " + str(n / (4 * channels * 64)))

  return data

def analyze_trx_tdd(data, channels):
  null_frames = 0
  tdd_switch_list = []
  prev_empty = False
  prev_i = 0
  n = len(data)
  frame_len = channels * 4 * 64

  max_iq_sample = 0
  min_iq_sample = 0
  avg_iq_sample = 0
  
  for i in range(int(n / frame_len)):
    iq_samples = data[i * frame_len:(i+1) * frame_len]
    iq_samples = [struct.unpack('f', iq_samples[4*j:4*(j+1)])[0] for j in range(int(len(iq_samples) / 4))]

    iq_samples_abs_avg = sum(map(abs, iq_samples))
    max_iq_sample = max(max_iq_sample, max(iq_samples))
    min_iq_sample = min(min_iq_sample, min(iq_samples))
    avg_iq_sample += iq_samples_abs_avg
  
    empty = iq_samples_abs_avg < 0.1
    if i == 0:
      first_tdd_period = empty
    if i > 0 and empty != prev_empty:
        tdd_switch_list.append(i-prev_i)
        prev_i = i
    null_frames += empty
    prev_empty = empty
  tdd_switch_list.append(i-prev_i)
  avg_iq_sample /= (n / frame_len) * (64 * channels)

  print('Max IQ sample: {}, Min IQ sample: {}'.format(max_iq_sample, min_iq_sample))
  print('Avg IQ sample: {}'.format(avg_iq_sample))
  print('Frame analyzed: ' + str(i))

  total_frames = int(n / frame_len)
  ratio = total_frames / null_frames if null_frames > 0 else 'inf'
  print('TDD ratio: {}'.format(ratio))
  print('TDD switch list: ' + ', '.join(map(str, tdd_switch_list)) + '\n')
  print('First TDD period: ' + "Not emitting" if first_tdd_period else "emitting")

BF1_PATH="../bf1/bf1"

def float_to_byte(float_list):
  iq_samples_bytes = bytearray(64 * 4)
  for i in range(64):
    b = struct.pack('f', float_list[i])
    for j in range(4):
      iq_samples_bytes[i * 4 + j] = b[j]
  return bytes(iq_samples_bytes)

def byte_to_float(b):
  float_list = []
  for i in range(64):
    x = struct.unpack('f', b[4 * i:4 * (i+1)])[0]
    float_list.append(x)
  return float_list

def compress(data):
  s = subprocess.run(
                  [BF1_PATH, "c"],
                  input=data,
                  capture_output=True,
  )
  return s.stdout

def decompress(data):
  s = subprocess.run([BF1_PATH, "d"],
                  input=data,
                  capture_output=True,
  )
  return s.stdout

def open_all(n):
  return (
    read_trace('tx.trace', n),
    read_trace('rx.trace', n),
    read_trx_trace('trxw.trace', n, 4),
    read_trx_trace('trxr.trace', n, 1),
  )

def analyze_all_tdd(n=0, data=None):

  if data:
    tx_data, rx_data, trxw_data, trxr_data = data
    print_header("TX TDD analysis")
    analyze_tdd(tx_data)
    print_header("RX TDD analysis")
    analyze_tdd(rx_data)
    print_header("TRXW TDD analysis")
    analyze_trx_tdd(trxw_data, 4)
    print_header("TRXR TDD analysis")
    analyze_trx_tdd(trxr_data, 1)
  else:
    data = read_trace('rx.trace', n)
    analyze_tdd(data)
    data = read_trace('tx.trace', n)
    analyze_tdd(data)
    data = read_trx_trace('trxw.trace', n, 4)
    analyze_trx_tdd(data, 4)
    data = read_trx_trace('trxr.trace', n, 1)
    analyze_trx_tdd(data, 1)

def check_tx_trxw(n=0, data=None):
  if data:
    tx_data, rx_data, trxw_data, trxr_data = data
  else:
    tx_data = read_trace('tx.trace', n)
    trxw_data = read_trx_trace('trxw.trace', n, 4)
  for i in range(int(min(len(tx_data) / 262, len(trxw_data) / (64 * 4 * 4)))):
    for j in range(1):
      k = i * 4 + j
      tx_iq = tx_data[(i * 262 + 22 + j * 60):(i * 262 + 22 + (j+1) * 60)]
      trxw_iq = trxw_data[(k * (64 * 4)):((k+1) * (64 * 4))]
      tx_iq_d = decompress(tx_iq)
      trxw_iq_d = compress(trxw_iq)
      if (_hash(tx_iq) != _hash(trxw_iq_d)) or (_hash(trxw_iq) != _hash(tx_iq_d)):
        print((i,j))
        print_hash("TX", tx_iq)
        print_hash("TRX compressed", trxw_iq_d)
        print_hash("TRX", trxw_iq)
        print_hash("TX decompressed", tx_iq_d)

def print_all(n=0, data=None):
  if data:
    tx_data, rx_data, trxw_data, trxr_data = data
    print_header('TX DATA')
    print_frame_list(tx_data, 0, 10)
    print_header('RX DATA')
    print_frame_list(rx_data, 0, 10)
    print_header('TRXW DATA')
    print_iq_list(trxw_data, 0, 10, 4)
    print_header('TRXR DATA')
    print_iq_list(trxr_data, 0, 10, 1)

data = open_all(1000)
#check_tx_trxw(data=data)
print_all(data=data)
analyze_all_tdd(data=data)
